# -*- coding: utf-8 -*-
from __future__ import unicode_literals, print_function

from .skipnose import SkipNose  # noqa


__author__ = 'Miroslav Shubernetskiy'
__email__ = 'miroslav@miki725.com'
__version__ = '0.3.0'
